
/* Marsaglia random number generator module */

#include "ranlaf.h"

#define CD 7654321L
#define CM (NO_NUMBERS-3)

#define FIRSTDIM 64
#define LASTDIM 97

static long x[LASTDIM], *j_ptr, *k_ptr, c;

void rmarin(short ij, short kl)
{
	short i, j, k, l, m, ii, jj;
	long ss,tt;

/*
	ij or kl too large will cause redundancy,
	negative will result in out-of-range i, j, k or l;

	former won't cause irand to return bad values but latter might, so
	make sure ij and kl are positive before proceeding
*/

	while (ij<0) ij+=31328;
	while (kl<0) kl+=30081;

/*
	compute i, j, k and l from ij and kl

	2 <= i < 179, 2 <= j < 179, 1 <= k < 179, 0 <= l < 169
*/

	i = ij / 177 % 177 + 2;
	j = ij % 177 + 2;
	k = kl / 169 % 178 + 1;
	l = kl % 169;

/* ii cycles through all elements of x, the random look-up table */

	for (ii = LASTDIM; ii--; )
	{

/*
	initialize ss (the bit accumulator)
	initialize tt (the single-bit mask)
	jj cycles through 24 bits
*/

		for (ss = 0, tt = NO_NUMBERS, jj = 1; jj <= 24; jj++)
		{

/* 0 <= m < 179 */

			m = i * j % 179 * k % 179;

/* shift j, k and m into i, j and k */

			i = j, j = k, k = m;

/* 0 <= l < 169 */

			l = (53 * l + 1) % 169;

/* tt shifts to next bit */

			tt >>= 1;

/* accumulate tt into ss, about half the time */

			if (l * m & 32) ss |= tt;
		}

/* put ss into look-up table */

		x[ii] = ss;
	}

/* assign remaining globals: j_ptr, k_ptr and c */

	j_ptr = x;
	k_ptr = x+FIRSTDIM;
	c = 362436L;
}

long irand(void)
{
	long uni;

/* uni is difference between two table values */

	uni = *j_ptr - *k_ptr;

/* make sure 0<=uni<NO_NUMBERS */

	if (uni < 0) uni += NO_NUMBERS;

/* assign uni into look-up table */
/* old *j_ptr = (new *j_ptr + *k_ptr) modulus NO_NUMBERS */

	*j_ptr = uni;

/* point j_ptr and k_ptr at new look-up table values */

	if ( ++j_ptr >= x + LASTDIM) j_ptr = x;
	if ( ++k_ptr >= x + LASTDIM) k_ptr = x;

/* modify c; 0<=c<NO_NUMBERS-CD or CM-CD<=c<CM */
/* there are three overlap positions between CM-CD and NO_NUMBERS-CD,
	 and three unavailable positions between CM and NO_NUMBERS */

	if ( (c -= CD) < 0) c += CM;

/* modify uni, make sure 0<=uni<NO_NUMBERS */

	if ( (uni -= c) < 0) uni += NO_NUMBERS;

	return uni;
}
/* uncomment following to run test: */

/*
#include <stdio.h>

static long correct[6]=
{
	6533892L,
	14220222L,
	7275067L,
	6172232L,
	8354498L,
	10633180L
};

void main(void)
{
	short i;
	long r;

	rmarin(1802, 9373);

	printf(
		"\n"
		"\n"
		"\"Correct\"   \"Actual\"\n"
		"---------   --------\n");

	for(i = 0; i < 20000; i++) (void)irand();

	for(i = 0; i < 6; i++)
	{
		r = irand();
		printf("  %6lx     %6lx    %s\n",
			correct[i], r,
			correct[i] == r ? "OK" : "BAD");
	}
}
*/
